--[[

    File: hsv.lua 

    Copyright (C) 2000-2009 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

--]]

require 'common.util'
require 'common.tex'

local huepts = {
	{1,0,0},
	{1,1,0},
	{0,1,0},
	{0,1,1},
	{0,0,1},
	{1,0,1}
}

function v3lerp(src,dst,t)
	local it = 1 - t
	local res = {}
	for i=1,#src do
		res[i] = src[i] * it + dst[i] * t
	end
	return res
end

function hsv2rgb(hsv)
	local r, g, b;
	local i, f, p, q, t;

	local h = (hsv[1] % 1) * #huepts
	local s = math.clamp(hsv[2], 0, 1)	-- s interpolates the value from grey
	local v = math.clamp(hsv[3], 0, 1)	-- v interpolates the value from black
	
	local ih = math.floor(h)
	local fh = h - ih
	ih = ih + 1
	local nh = (ih % #huepts) + 1
	local hc = v3lerp( huepts[ih], huepts[nh], fh)
	local hsc = v3lerp({1,1,1}, hc, s)
	return v3lerp({0,0,0}, hsc, v)
end

HsvTex = class(Tex1D)
function HsvTex:init(detail)
	HsvTex.super.init(self, {
		minFilter=gl.GL_NEAREST;
		magFilter=gl.GL_NEAREST;
		width=detail;
		internalFormat=gl.GL_RGB;
		data=function(u,v)
			local v = hsv2rgb{u / detail, 1, 1}
			return v[1], v[2], v[3], 1	
		end;
	})
end

