--require 'util'

function exec(str, chunk)
	local fn, errmsg
	fn, errmsg = loadstring(str, chunk)
	if not fn then
		print('ERROR:'..tostring(errmsg))
		print('CHUNK:'..chunk)
		print('CODE:\n'..str)
		assert(false, 'ERROR - loadstring ' .. chunk .. ' failed!')
	end
	fn()
end


function addfunc(class,name, fname,codeFromParamsToReturn)
	exec('temp = function' .. codeFromParamsToReturn .. '\nend\n', name..' '..fname)
	class[fname] = temp
	temp = nil
end

function arrayref(n)
	if n then
		return function(i) return n..'['..i..']' end
	else
		return function(i) return i end
	end
end

function arrayop(dim, ...)
	local fs = {...}
	local str = ''
	for i=1,dim do
		for j,f in ipairs(fs) do
			if type(f) == 'string' then
				str = str .. f
			elseif type(f) == 'function' then
				str = str .. f(i)
			end
		end
		str = str .. '\n'
	end
	return str
end

function addvecop(class,dim,fname,op)
	-- a = b <op> c, return a
	local str = '(a,b,c)\n'
	str = str .. 'if type(c) == "table" then\n'
	str = str .. arrayop(dim, arrayref('a'), '=', arrayref('b'), op, arrayref('c'))
	str = str .. 'else\n'
	str = str .. arrayop(dim, arrayref('a'), '=', arrayref('b'), op, 'c')
	str = str .. 'end\n'
	str = str .. 'return a\n'
	addfunc(class, 'vec'..dim, fname, str)
	
	-- a = a <op> b, return a
	str = '(a,b)\n'
	str = str .. 'if type(b) == "table" then\n'
	str = str .. arrayop(dim, arrayref('a'), '=', arrayref('a'), op, arrayref('b'))
	str = str .. 'else\n'
	str = str .. arrayop(dim, arrayref('a'), '=', arrayref('a'), op, 'b')
	str = str .. 'end\n'
	str = str .. 'return a\n'
	addfunc(class, 'vec'..dim, fname..'To', str)

	str = [[(a,b)
local r = vec]]..dim..[[:new()
vec]]..dim..[[.]]..fname..[[(r, a, b)
return r
]]
	addfunc(class, 'vec'..dim, fname..'Ret', str)
	local mt = class.metatable
	mt['__'..fname] = class[fname..'Ret']

	temp = nil	-- clear it
end

function buildvecclass(dim)
	local class = {}
	local name = 'vec'..dim
	local mname = name..'Meta'
	
	-- register it globally
	_G[name] = class
	-- and its meta table
	local metatable = { __index = class }
	class.name = name
	class.metatable = metatable
	_G[mname] = metatable
	
	local str = [[
(vecclass,...)
local args = {...}
local v = {]]
	local comma = ''
	for i=1,dim do 
		str = str .. comma .. '0'
		comma = ','
	end
	str = str .. [[}
setmetatable(v,]]..mname..[[)
if #args == 1 then
local arg = args[1]
if type(arg) == "table" then
]] .. arrayop(dim, arrayref('v'), '=', arrayref('arg')) .. [[
else
]] .. arrayop(dim, arrayref('v'), '=', 'arg') .. [[
end
elseif #args == ]]..dim..[[ then
]] .. arrayop(dim, arrayref('v'), '=', arrayref('args')) .. [[
end
return v
]]
	addfunc(class, name, 'new', str)	
	addfunc(class, name,'set', '(v,s)\n' .. arrayop(dim, arrayref('v'), '= s') .. 'return v\n')
	
	str = '(v)\n'
	str = str .. 'return ""'
	comma = ''
	for i=1,dim do
		str = str .. comma .. '..v['..i..']'
		comma = '..","'
	end
	addfunc(class, name, 'tostring', str)

	str = '(a,b) return '
	comma = ''
	for i=1,dim do
		str = str .. comma .. 'a['..i..']*b['..i..']'
		comma = ' + '
	end
	addfunc(class, name, 'dot', str)
	
	addfunc(class, name, 'lenSq', '(v) return '..name..'.dot(v,v)')
	addfunc(class, name, 'length', '(v) return math.sqrt('..name..'.dot(v,v))')
	addfunc(class, name, 'normalize', '(v) '..name..'.mulTo(v, 1/'..name..'.length(v)) return v')
	
	addvecop(class, dim, 'add', '+')
	addvecop(class, dim, 'sub', '-')
	addvecop(class, dim, 'mul', '*')
	addvecop(class, dim, 'div', '/')
	
end

buildvecclass(2)
buildvecclass(3)

vec3.mulAdd = function(a,b,c,d)
	d[1] = a[1] + b[1] * c
	d[2] = a[2] + b[2] * c
	d[3] = a[3] + b[3] * c
	return d
end

vec3.mulAddTo = function(a,b,c)
	a[1] = a[1] + b[1] * c
	a[2] = a[2] + b[2] * c
	a[3] = a[3] + b[3] * c
	return a
end

vec3.cross = function(a,b)
	return {
		a[2]*b[3] - a[3]*b[2],
		a[3]*b[1] - a[1]*b[3],
		a[1]*b[2] - a[2]*b[1]
	}
end

vec3.toQuat = function(src,dst)
	--todo - instead of determining the dot product,
	--determine it from the cross product magnitude
	--when you go to normalize the axis
	local cosom = vec3.dot(src,dst)
	if cosom < -1 then cosom = -1 end
	if cosom > 1 then cosom = 1 end

--	return vec4f(vec3unit(src * dst) * sinhalfom, (float)cos(halfom))

	local axis = vec3.cross(src,dst)
	local lenSq = vec3.lenSq(axis)	--get the sin(theta) length
	if lenSq <= 0.000001 then	--if we have close to no movement
		return {0,0,0,1}	--just return the identity
	end --else...

	local halfom = math.acos(cosom) * 0.5	-- * angleScale
	local sinhalfom = math.sin(halfom)
	local scale = sinhalfom / math.sqrt(lenSq)	-- check for div by zero?
	return {
		axis[1] * scale,
		axis[2] * scale,
		axis[3] * scale,
		math.cos(halfom)
	}
end

buildvecclass(4)
